/*!
 * \file      http_layer.hh
 * \brief     Header file for ITS HTTP protocol layer.
 * \author    ETSI STF549
 * \copyright ETSI Copyright Notification
 *            No part may be reproduced except as authorized by written permission.
 *            The copyright and the foregoing restriction extend to reproduction in all media.
 *            All rights reserved.
 * \version   0.1
 */
#pragma once

#include <memory>

#include "t_layer.hh"

#include "http_codec.hh"

namespace LibItsHttp__TypesAndValues {
  class HeaderLines; //! Forward declaration of TITAN class
}

class OCTETSTRING; //! Forward declaration of TITAN class

/*!
 * \class http_layer
 * \brief  This class provides a factory class to create an udp_layer class instance
 */
class http_layer : public layer {
  params _params;
  http_codec _codec;
  std::unique_ptr<LibItsHttp__TypesAndValues::HeaderLines> _headers;

public: //! \publicsection
  /*!
   * \brief Specialised constructor
   *        Create a new instance of the http_layer class
   * \param[in] p_type \todo
   * \param[in] p_param \todo
   */
  http_layer() : layer(), _params(), _headers(nullptr) {};
  /*!
   * \brief Specialised constructor
   *        Create a new instance of the http_layer class
   * \param[in] p_type \todo
   * \param[in] p_param \todo
   */
  http_layer(const std::string& p_type, const std::string& p_param);
  /*!
   * \brief Default destructor
   */
  virtual ~http_layer() { _headers.reset(nullptr); };

  /*!
   * \virtual
   * \fn void send_data(OCTETSTRING& data, params& params);
   * \brief Send bytes formated data to the lower layers
   * \param[in] p_data The data to be sent
   * \param[in] p_params Some parameters to overwrite default value of the lower layers parameters
   */
  virtual void send_data(OCTETSTRING& data, params& params);
  /*!
   * \virtual
   * \fn void receive_data(OCTETSTRING& data, params& params);
   * \brief Receive bytes formated data from the lower layers
   * \param[in] p_data The bytes formated data received
   * \param[in] p_params Some lower layers parameters values when data was received
   */
  virtual void receive_data(OCTETSTRING& data, params& info);
}; // End of class http_layer

